package Desktop::DBI;

require Exporter;
no encoding;
use utf8;

@ISA = qw(Exporter);

%EXPORT_TAGS = ( 'all' => [ 
 qw(queryA queryH query
    iterateA iterateH) ] );

@EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );
@EXPORT = ( @{ $EXPORT_TAGS{'all'} } );
$VERSION = '0.6';

use strict;
use warnings;

use strict;
use vars qw($VERSION);
use Carp;
use DBI;
use Desktop::P;

$VERSION = '0.93';

BEGIN {
  eval {
    require Encode::compat if $] < 5.007001;
    require Encode;
  }
}


# Default options for the module
my %DefaultOptions = (
		      'user'       =>  '',
		      'password'   =>  '',
		      'AUTOCOMMIT' =>  1,
		      'WARN'       =>  0,
		      'DEBUG'      =>  0,
		      'CLOBBER'    =>  0,
		      'CASESENSITIV' => 0,
		      );

use Data::Dumper;

our $my_db;
our $my_rows;

################################################################
# connect interface
sub connect {
  my $db = DBI->connect("dbi:Pg:",$ENV{'PGUSER'},$ENV{'PGPASSWORD'},
#			{ RaiseError => 1, AutoCommit => 0 });
			{ RaiseError => 0, AutoCommit => 1 });
  if(!$db) {
    die "
Can't connect to the database.
Perhaps you need to set PGDATABASE and PGUSER
environment variables.
";
  }

  $my_db=$db;
  return $db;
}

sub commit1 {
  $my_db->commit;
}

# TIEHASH interface
# tie %h,Tie::DBI,[dsn|dbh,table,key],\%options
sub TIEHASH {
    my $class = shift;
    my $info = shift;
    my ($dbh,$table,$key,$opt) = ($info->{db},$info->{table},$info->{key},$info);

    $dbh = coalesce($dbh,$my_db);
    croak "Usage tie(%h, 'Desktop::DBI', db, table,key, \\%options)\n" 
      unless $dbh && $table && $key;

    my $self = {
	%DefaultOptions,
	defined($opt) ? %$opt : ()
	};
    bless $self,$class;
    my ($driver);

    if (UNIVERSAL::isa($dbh,'DBI::db')) {
      $dbh->{Warn} = $self->{WARN};
    } else {
      croak "TIEHASH: Can't open database, ",$class->errstr unless $dbh;
    }

    # set up more instance variables
    @{$self}{'dbh','table','key','driver','info'} = ($dbh, $table, $key, $driver,$info);

    return $self;
}


sub FETCH {
    my($s,$key) = @_;

    my $h = {};
    if(tie %$h,'Desktop::DBI::Record',$s,$key) {
      return $h;
    } else {
      return undef;
    }
}

sub STORE {
    my ($table,$key,$value) = @_;
    # There are two cases where this can be called.  In the first case, we are
    # passed a hash reference to field names and their values.  In the second
    # case, we are passed a Tie::DBI::Record, for the purposes of a cloning
    # operation.
    croak "STORE: attempt to store non-hash value into record"
	unless ref($value) eq 'HASH';

    my @keys1;
    my @values;
    my @v1;

    if(ref($table->{key})) {
      # composite key defined
      if(ref($key) eq 'ARRAY') {
	# composite key used
	for my $k (@{$key}) {
	  for my $i (0..$#{$key}) {
	    $value->{$table->{key}[$i]}=$key->[$i];
	  }
	}
      } else {
	# single key used
	croak "TIEHASH: non array used on composite key definition";
      }
    } else {
      # single key defined
      if(ref($key)) {
	croak "TIEHASH: composite key used on single key definition";
      } else {
	# single key used
	$value->{$table->{key}}=$key;
      }
    }


    while (my ($key1, $val) = each %{$value}) {
      push @keys1,Desktop::DBI::Record::sqlqi($key1);
      push @values,$val;
      push @v1,"?";
    }

    my $query=
      "INSERT INTO ".Desktop::DBI::Record::sqlqi($table->{table})." ".
      " (".join(",",@keys1).") ".
      "VALUES (".join(",",@v1).")";

    unless($table->_query($query,@values)) {
      if($table->{CROAK}) {
	croak "STORE: insert failed";
      } else {
	warn "STORE: insert failed";
      }
    }

    # Attempt an insert.  If that fails (usually because the key already exists),
    # perform an update. For this to work correctly, the key field MUST be marked unique


    # Neat special case: If we are passed an empty anonymous hash, then
    # we must tie it to Tie::DBI::Record so that the correct field updating
    # behavior springs into existence.
    if(tie %$value,'Desktop::DBI::Record',$table,$key) {
      return $value;
    } else {
      return undef;
    }
}

sub autodetect_keys {
  my $s=shift;

  my $sth=$s->_query("
    SELECT column_name
    FROM information_schema.key_column_usage
    WHERE table_name = ?
    ORDER BY ordinal_position",$s->{table});
}

sub _query {
  my $me=shift;
  my ($query,@param)=@_;

  if($me->{info}{echo}) {
    print "QUERY:$query [".join(",",@param)."]\n";
  }

  my $db=$me->{dbh};
  my $sth=$db->prepare_cached($query);

  $sth->execute(@param);
  my $rows=$sth->rows;
  if($db->errstr) {
    warn 
      $query."[".join(",",@param)."]\n".
      $db->errstr."\n";
  } else {
    if(wantarray) {
      my $r = $sth->fetchrow_arrayref;
      if($r) {
	return @$r;
      } else {
	return undef;
      }
    } else {
      return $sth;
    }
  }
}

package Desktop::DBI::Record;
use strict;
use vars qw($VERSION);
use Carp;
use DBI;
$VERSION = '0.50';

use Data::Dumper;

# TIEHASH interface
# tie %h,Tie::DBI::Record,dbh,table,record
sub TIEHASH {
    my $class = shift;
    my ($table,$key) = @_;
    my $me = bless { 
	'table'=>$table,   # table object
	'record'=>$key, # the record we're keyed to
	},$class;

    if(ref($table->{key})) {
      # composite key defined
      if(ref($key) eq 'ARRAY') {
	# composite key used
	$me->{keys1}=$key;
	$me->{where} = "( ".join(" AND ",(map {" $_ = ? "} @{$table->{key}}))." )";
      } else {
	# single key used
	croak "TIEHASH: non array used on composite key definition";
      }
    } else {
      # single key defined
      if(ref($key)) {
	croak "TIEHASH: composite key used on single key definition";
      } else {
	# single key used
	$me->{keys1}=[$key];
	$me->{where} = "$table->{key} = ?";
      }
    }

    $me->_load();
#    print "ROWS:$me->{rows}\n";

    unless($me->{rows}>=1) { return undef; }
    return $me;
}

sub _load {
  my ($me)=@_;

  if(!$me->{where}) { croak "_LOAD:no where"; }
  if(!$me->{keys1}) { croak "_LOAD:no keys"; }

  my $query="SELECT * FROM ".sqlqi($me->{table}{table})." WHERE $me->{where}";

#  print Dumper($me->{keys1});

  my $sth=$me->{table}->_query($query,@{$me->{keys1}});
  my $rec=$sth->fetchrow_hashref;
  $me->{rows}=$sth->rows;

  $me->{rec}=$rec;

  $sth->finish;

#  print "_LOAD:".Dumper($rec);
}

sub FETCH {
    my ($s,$field) = @_;
    return undef unless $s->{'rec'};
    return $s->{'rec'}{$field};
}

sub DELETE1 {
    my ($s,$field) = @_;
    $s->STORE($field,undef);
}

sub STORE {
    my ($s,$field,$value) = @_;
    my $query;

    if($s->{rec}) {
      $query=
	"UPDATE ".sqlqi($s->{table}{table}).
	" SET ".sqlqi($field)." = ? WHERE ".$s->{where};
    } else {
    }

    $s->{table}->_query($query,$value,@{$s->{keys1}});
}

# Can't delete the record in this way, but we can
# clear out all the fields by setting them to undef.
sub CLEAR1 {
    my ($s) = @_;
    croak "CLEAR: read-only database"
	unless $s->{'table'}->{CLOBBER} > 1;
    my %h = map { $_ => undef} keys %{$s->{'table'}->_fields()};
    delete $h{$s->{'record'}};  # can't remove key field
    $s->{'table'}->STORE($s->{'record'},\%h);
}

sub FIRSTKEY {
    my $s = shift;
    each %{$s->{'rec'}};
}

sub NEXTKEY {
    my $s = shift;
    each %{$s->{'rec'}};
}

sub EXISTS1 {
    my $s = shift;
    return $s->{'table'}->_fields()->{$_[0]};
}

sub DESTROY1 {
    my $s = shift;
    warn "$s->{table}:$s->{value} has been destroyed" if $s->{'table'}->{DEBUG};
}

sub sqlqi {
  my ($val)=@_;

  $val=~s/"/""/g;
  $val=qq{"$val"};

  return $val;
}


package Desktop::DBI;

################################################################

sub iterateH {
  my ($sth,$func)=@_;

  unless(ref($sth) eq 'DBI::st') { return undef; }

  while(my $rec=$sth->fetchrow_hashref()) {
    &{$func}($rec);
  }
}

sub iterateA {
  my ($sth,$func)=@_;

  unless(ref($sth) eq 'DBI::st') { return undef; }

  while(my $rec=$sth->fetchrow_arrayref()) {
    &{$func}($rec);
  }
}

sub query { queryA(@_); }

sub queryA {
  my ($query,@param)=@_;

  my $sth=$my_db->prepare($query);
  $sth->execute(@param);
  $my_rows=$sth->rows;
  if($my_db->errstr) {
    warn 
      $query."[".join(",",@param)."]\n".
      $my_db->errstr."\n";
  } else {
    if(wantarray) {
      my $r = $sth->fetchrow_arrayref;
      if($r) {
	return @$r;
      } else {
	return undef;
      }
    } else {
      return $sth;
    }
  }
}

sub queryH {
  my ($query,@param)=@_;

  my $sth=$my_db->prepare($query);
  $sth->execute(@param);
  $my_rows=$sth->rows;
  if($my_db->errstr) {
    warn 
      $query."[".join(",",@param)."]\n".
      $my_db->errstr."\n";
  } else {
    return $sth->fetchrow_hashref;
  }
}
################################################################


1;

=POD
################################################################

We need to fix Tie::DBI to support composite primary keys.
Examples:

# old way
tie %h,Tie::DBI,{table=>'globs', key=>'type'};

# autodetect primary key if not specified
tie %h,Desktop::DBI,{table=>'globs'};

# specify composite primary key by using an array reference
tie %h,Desktop::DBI,{table=>'type-mime', key=>['glob','type']}; 

# BUG: handle table-name quoting for funny table names

# data access for composite keys

print $h{['*.c',undef]};
print $h{{glob=>'*.gif',type=>'image/gif'}};

